import { __awaiter } from "tslib";
import { useCallback, useEffect, useRef, useState } from 'react';
import { isPromiseLike } from 'retil-support';
/**
 * The operation function itself is passed in as an argument, so that we're able
 * to type the output parameter.
 *
 * Note that if you'd like to execute an arbitrary function in response to an
 * event, you can make an operation whose input is itself an async function.
 */
export function useOperation(operation, deps) {
    // The types don't allow an `undefined` value to be passed in, but it works.
    //@ts-ignore
    // eslint-disable-next-line react-hooks/exhaustive-deps
    const memoizedOperation = useCallback(operation, deps);
    const [error, setError] = useState([]);
    const [state, setState] = useState([false]);
    const abortControllerRef = useRef(null);
    const mountedRef = useRef(true);
    useEffect(() => {
        mountedRef.current = true;
        return () => {
            mountedRef.current = false;
            if (abortControllerRef.current) {
                abortControllerRef.current.abort();
                abortControllerRef.current = null;
            }
        };
    }, []);
    const trigger = useCallback((data) => __awaiter(this, void 0, void 0, function* () {
        try {
            if (!mountedRef.current) {
                return Promise.reject();
            }
            if (abortControllerRef.current) {
                abortControllerRef.current.abort();
            }
            const abortController = new AbortController();
            const signal = abortController.signal;
            abortControllerRef.current = abortController;
            const resultPromise = memoizedOperation(data, signal);
            if (resultPromise && isPromiseLike(resultPromise)) {
                const instantProbe = Symbol();
                const probeResult = yield Promise.race([
                    resultPromise,
                    Promise.resolve(instantProbe),
                ]);
                if (probeResult !== instantProbe) {
                    // If the result promise resolves before a newly created promise, then
                    // let's skip the pending state and immediately set the result on state.
                    if (abortControllerRef.current === abortController) {
                        setState([false, probeResult]);
                    }
                    return probeResult;
                }
                else {
                    if (abortControllerRef.current === abortController) {
                        setState([true]);
                    }
                    const result = yield resultPromise;
                    if (abortControllerRef.current === abortController) {
                        abortControllerRef.current = null;
                        setState([false, result]);
                    }
                    return result;
                }
            }
            else {
                setState([false, resultPromise]);
                return resultPromise;
            }
        }
        catch (error) {
            setError([error]);
            return Promise.reject();
        }
    }), [memoizedOperation]);
    if (error.length) {
        throw error[0];
    }
    return [trigger].concat(state);
}
